/*
 * This program is to solve the system of nonlinear equations
 * $$
 *	\left\{
 *		\begin{array}{l}
 *	 		x^2 - y     = 0 \\
 * 			-sin(x) + y = 0
 * 		\end{array}
 * 	\right.
 * 	$$
 * by means of Newthon-Raphson method.
 *
 * (C) vladimir.smirnov.penza@gmail.com, 2010
 * Redistriution in source or binary form, modified or not, permitted
 * under the terms of BSD license.
 *
 * The latest version of this file can be obtained from:
 * ftp://ftp.pguas.ru/pub/math/edu-examples/nr2lmr2.tar.gz
 *
 * To compile & run:
 *  - Windows: install MinGW from http://sourceforge.net/projects/mingw,
 *	than say 'make'.
 *  - GNU environment (Linux, notably): say 'make'.
 *  - generic UN*X: _try_ to say 'gmake' (you'll need GCC and GNU make).
 */

#include "../common.h"
#include "../common.c"

static vboolean
NewtonRaphson(vboolean verbose,vec_t lambda)
{
	int i;
	vboolean	last_iter = false;
	matrix2_t	J, Jinv;
	vec2_t		F, X = FIRST_APPROX, Xnext;
	/* ---------------- The method itself, with all checks ------------ */
	for(i = 0;i < MAX_ITERATIONS;i++)
	{
		getF(F, X);
		if(verbose)
		{
			printf("\nIteration: %i; X: ", i);
			vec2Dump(X, "X[1] = ", "\tX[2] = ", "\t");
			printf("Modulus of residual %g\n", vec2Abs(F));
		}
		else
		{
			vec2Dump(X, "", ";", ";");
			printf("%g\n", vec2Abs(F));
		}
		if(last_iter)
			break;
		getJ(J, X);
		if(verbose)
		{
			vec2Dump(F, "F[1] = ", "\tF[2] = ", "\n");
			mat2Dump(J, Jlabels);
		}
		if(!mat2Inv(Jinv, J))
		{
			printf("STOP: Jacobian reached zero.\n");
			return false;
		}
		if(verbose)
			mat2Dump(Jinv, Jinvlabels);
		mat2VecMul(Xnext, Jinv, F);
		vec2Scale(Xnext, Xnext, lambda);
		vec2Sub(Xnext, X, Xnext);
		if(isAway(X, Xnext))
		{
			printf("STOP: divergence of iteration process.\n");
			return false;
		}
		if(isNear(X,Xnext,DESIRED_ACCURACY_SQUARED))
			last_iter = true;
		vec2Copy(X,Xnext);
	}
	if(!last_iter)
	{
		printf("Can't reach solution after %i iterations.\n", i);
		return false;
	}
	/* ---------------------------------------------------------------- */
	printf("\nSolution reached after %i iterations\n", i);
	return true;
}

int
main(int argc, char *argv[])
{
	int retcode = -1;
	vboolean verbose = false;
	vec_t lambda;
	char *endptr = NULL;
	if(argc < 2)
	{
		printf("Usage: %s <float lambda> [--verbose]\n", argv[0]);
		goto bye;
	}
	lambda = strtod(argv[1], &endptr);
	if(!endptr || *endptr)
	{
		printf("Second parameter must be valid floating-point value.\n");
		goto bye;
	}
	if(lambda < FLOAT_ZERO_EPSILON)
	{
		printf("\"lambda\" too small to converge.\n");
		goto bye;
	}
	if(argc > 2 && !strcmp(argv[2],"--verbose"))
		verbose = true;
	retcode = NewtonRaphson(verbose, lambda) ? 0 : -1;
bye:;
	return retcode;
}
